#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.


import time
import os.path
import ctypes
import logging
import libarchive
import collections
import stat

from diffoscope.exc import ContainerExtractionError
from diffoscope.config import Config
from diffoscope.excludes import any_excluded
from diffoscope.tempfiles import get_temporary_directory
from diffoscope.difference import Difference

from ..device import Device
from ..symlink import Symlink
from ..directory import Directory
from ..socket_or_fifo import SocketOrFIFO

from .archive import Archive, ArchiveMember

logger = logging.getLogger(__name__)


# Monkeypatch libarchive-c (<< 2.2)
if not hasattr(libarchive.ffi, "entry_rdevmajor"):
    libarchive.ffi.ffi(
        "entry_rdevmajor", [libarchive.ffi.c_archive_entry_p], ctypes.c_uint
    )
    libarchive.ArchiveEntry.rdevmajor = property(
        lambda self: libarchive.ffi.entry_rdevmajor(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_rdevminor"):
    libarchive.ffi.ffi(
        "entry_rdevminor", [libarchive.ffi.c_archive_entry_p], ctypes.c_uint
    )
    libarchive.ArchiveEntry.rdevminor = property(
        lambda self: libarchive.ffi.entry_rdevminor(self._entry_p)
    )
# Monkeypatch libarchive-c (<< 2.3)
if not hasattr(libarchive.ffi, "entry_nlink"):
    libarchive.ffi.ffi(
        "entry_nlink", [libarchive.ffi.c_archive_entry_p], ctypes.c_uint
    )
    libarchive.ArchiveEntry.nlink = property(
        lambda self: libarchive.ffi.entry_nlink(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_uid"):
    libarchive.ffi.ffi(
        "entry_uid", [libarchive.ffi.c_archive_entry_p], ctypes.c_uint32
    )
    libarchive.ArchiveEntry.uid = property(
        lambda self: libarchive.ffi.entry_uid(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_gid"):
    libarchive.ffi.ffi(
        "entry_gid", [libarchive.ffi.c_archive_entry_p], ctypes.c_uint32
    )
    libarchive.ArchiveEntry.gid = property(
        lambda self: libarchive.ffi.entry_uid(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_mtime_nsec"):
    libarchive.ffi.ffi(
        "entry_mtime_nsec", [libarchive.ffi.c_archive_entry_p], ctypes.c_long
    )
    libarchive.ArchiveEntry.mtime_nsec = property(
        lambda self: libarchive.ffi.entry_mtime_nsec(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_uname"):
    libarchive.ffi.ffi(
        "entry_uname", [libarchive.ffi.c_archive_entry_p], ctypes.c_char_p
    )
    libarchive.ArchiveEntry.uname = property(
        lambda self: libarchive.ffi.entry_uname(self._entry_p)
    )
if not hasattr(libarchive.ffi, "entry_gname"):
    libarchive.ffi.ffi(
        "entry_gname", [libarchive.ffi.c_archive_entry_p], ctypes.c_char_p
    )
    libarchive.ArchiveEntry.gname = property(
        lambda self: libarchive.ffi.entry_gname(self._entry_p)
    )
# Monkeypatch libarchive-c (>= 2.8)
# Wire mtime_nsec attribute as some libarchive versions (>=2.8) don't expose it
# for ArchiveEntry. Doing this allows a unified API no matter which version is
# available.
if not hasattr(libarchive.ArchiveEntry, "mtime_nsec") and hasattr(
    libarchive.ffi, "entry_mtime_nsec"
):
    libarchive.ArchiveEntry.mtime_nsec = property(
        lambda self: libarchive.ffi.entry_mtime_nsec(self._entry_p)
    )


# Monkeypatch libarchive-c so we always get pathname as (Unicode) str
# Otherwise, we'll get sometimes str and sometimes bytes and always pain.
libarchive.ArchiveEntry.pathname = property(
    lambda self: libarchive.ffi.entry_pathname(self._entry_p).decode(
        "utf-8", errors="surrogateescape"
    )
)


def list_libarchive(path, ignore_errors=False):
    try:
        with libarchive.file_reader(path) as archive:
            for entry in archive:
                name_and_link = entry.name
                if entry.issym:
                    name_and_link = "{entry.name} -> {entry.linkname}".format(
                        entry=entry
                    )
                if Config().exclude_directory_metadata == "recursive":
                    yield "{name_and_link}\n".format(
                        name_and_link=name_and_link
                    )
                    continue
                if entry.isblk or entry.ischr:
                    size_or_dev = "{major:>3},{minor:>3}".format(
                        major=entry.rdevmajor, minor=entry.rdevminor
                    )
                else:
                    size_or_dev = "-" if entry.size is None else entry.size
                mtime = time.strftime(
                    "%Y-%m-%d %H:%M:%S", time.gmtime(entry.mtime)
                ) + ".{:06d}".format(entry.mtime_nsec // 1000)
                if entry.uname:
                    user = "{user:<8} {uid:>7}".format(
                        user=entry.uname.decode(
                            "utf-8", errors="surrogateescape"
                        ),
                        uid="({})".format(entry.uid),
                    )
                else:
                    user = entry.uid
                if entry.gname:
                    group = "{group:<8} {gid:>7}".format(
                        group=entry.gname.decode(
                            "utf-8", errors="surrogateescape"
                        ),
                        gid="({})".format(entry.gid),
                    )
                else:
                    group = entry.gid
                yield "{strmode} {entry.nlink:>3} {user:>8} {group:>8} {size_or_dev:>8} {mtime:>8} {name_and_link}\n".format(
                    strmode=entry.strmode.decode("us-ascii"),
                    entry=entry,
                    user=user,
                    group=group,
                    size_or_dev=size_or_dev,
                    mtime=mtime,
                    name_and_link=name_and_link,
                )
    except libarchive.exception.ArchiveError:
        if not ignore_errors:
            raise


class LibarchiveMember(ArchiveMember):
    def __init__(self, archive, entry):
        super().__init__(archive, entry.pathname)

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def is_socket_or_fifo(self):
        return False


class LibarchiveDirectory(Directory, LibarchiveMember):
    def __init__(self, archive, entry):
        LibarchiveMember.__init__(self, archive, entry)

    def compare(self, other, source=None):
        return None

    def has_same_content_as(self, other):
        return False

    @property
    def path(self):
        raise NotImplementedError(
            "LibarchiveDirectory is not meant to be extracted."
        )

    def is_directory(self):
        return True

    def get_member_names(self):
        raise ValueError("archives are compared as a whole.")  # noqa

    def get_member(self, member_name):
        raise ValueError("archives are compared as a whole.")  # noqa


class LibarchiveSymlink(Symlink, LibarchiveMember):
    def __init__(self, archive, entry):
        LibarchiveMember.__init__(self, archive, entry)
        self._destination = entry.linkpath

    @property
    def symlink_destination(self):
        return self._destination

    def is_symlink(self):
        return True


class LibarchiveDevice(Device, LibarchiveMember):
    def __init__(self, container, entry):
        LibarchiveMember.__init__(self, container, entry)
        self._mode = entry.mode
        self._major = entry.rdevmajor
        self._minor = entry.rdevminor

    def get_device(self):
        return (self._mode, self._major, self._minor)

    def is_device(self):
        return True


class LibarchiveFIFO(SocketOrFIFO, LibarchiveMember):
    def __init__(self, container, entry):
        LibarchiveMember.__init__(self, container, entry)
        self._mode = entry.mode

    def get_type(self):
        return stat.S_IFMT(self._mode)

    def is_socket_or_fifo(self):
        return True


class LibarchiveContainer(Archive):
    def open_archive(self):
        # libarchive is very very stream oriented an not for random access
        # so we are going to reopen the archive everytime
        # not nice, but it'll work
        return True

    def close_archive(self):
        pass

    def get_member_names(self):
        self.ensure_unpacked()
        return self._members.keys()

    def get_member(self, member_name):
        with libarchive.file_reader(self.source.path) as archive:
            for entry in archive:
                if entry.pathname == member_name:
                    return self.get_subclass(entry)
        raise KeyError(f"{member_name} not found in archive")

    def get_filtered_members(self):
        try:
            with libarchive.file_reader(self.source.path) as archive:
                for entry in archive:
                    if any_excluded(entry.pathname):
                        continue
                    yield entry.pathname, self.get_subclass(entry)
        except libarchive.exception.ArchiveError:
            pass

    def extract(self, member_name, dest_dir):
        self.ensure_unpacked()
        return self._members[member_name]

    def get_subclass(self, entry):
        if entry.isdir:
            return LibarchiveDirectory(self, entry)
        if entry.issym:
            return LibarchiveSymlink(self, entry)
        if entry.isblk or entry.ischr:
            return LibarchiveDevice(self, entry)
        if entry.isfifo:
            return LibarchiveFIFO(self, entry)

        return LibarchiveMember(self, entry)

    def ensure_unpacked(self):
        if hasattr(self, "_members"):
            return

        self._tmpdir_object = get_temporary_directory(
            suffix=self.__class__.__name__
        )
        tmpdir = self._tmpdir_object.name
        self._members = collections.OrderedDict()

        logger.debug("Extracting %s to %s", self.source.path, tmpdir)

        with libarchive.file_reader(self.source.path) as archive:
            for idx, entry in enumerate(archive):
                # Always skip directories
                if entry.isdir:
                    continue

                # Don't extract excluded files
                if any_excluded(entry.pathname):
                    continue

                # Keep directory sizes small. could be improved but should be
                # good enough for "ordinary" large archives.
                dst = os.path.join(tmpdir, str(idx // 4096), str(idx % 4096))
                _, ext = os.path.splitext(entry.pathname)
                dst += ext

                # Maintain a mapping of archive path to the extracted path,
                # avoiding the need to sanitise filenames.
                self._members[entry.pathname] = dst

                logger.debug("Extracting %s to %s", entry.pathname, dst)

                os.makedirs(os.path.dirname(dst), exist_ok=True)
                try:
                    with open(dst, "wb") as f:
                        for block in entry.get_blocks(block_size=2**17):
                            f.write(block)
                except Exception as e:
                    raise ContainerExtractionError(entry.pathname, e)

        logger.debug(
            "Extracted %d entries from %s to %s",
            len(self._members),
            self.source.path,
            tmpdir,
        )

    def comparisons(self, other):
        def hide_trivial_dirs(item):
            file1, file2, comment = item
            return not (
                isinstance(file1, Directory)
                and isinstance(file2, Directory)
                and comment is None
            )

        return filter(hide_trivial_dirs, super().comparisons(other))


class LibarchiveContainerWithFilelist(LibarchiveContainer):
    def compare(self, other, **kwargs):
        differences = []

        # Always include the filelist
        if other.source:
            differences = [
                Difference.from_text_readers(
                    list_libarchive(self.source.path),
                    list_libarchive(other.source.path),
                    self.source.path,
                    other.source.path,
                    source="file list",
                )
            ]

        differences.extend(super().compare(other, **kwargs))

        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import abc
import logging

from diffoscope.profiling import profile
from diffoscope.tempfiles import get_temporary_directory

from ..missing_file import MissingFile, AbstractMissingType

from .file import File
from .container import Container

logger = logging.getLogger(__name__)


class Archive(Container, metaclass=abc.ABCMeta):
    def __new__(cls, source, *args, **kwargs):
        if isinstance(source, MissingFile):
            return super(Container, MissingArchive).__new__(MissingArchive)
        return super(Container, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        with profile("open_archive", self):
            self._archive = self.open_archive()
        self._opened = True

    def __del__(self):
        """
        __del__ can be called even if __init__ did not fully complete. This
        becomes an issue if an `open_archive` call fails (due to a faulty
        archive, missing dependency, etc.), and close_archive then assumes that
        it was opened successfully.

        We therefore track whether open_archive actually worked, and only call
        close_archive if that was the case. (#276)
        """

        if hasattr(self, "_opened"):
            with profile("close_archive", self):
                self.close_archive()

    @property
    def archive(self):
        return self._archive

    @abc.abstractmethod
    def open_archive(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def close_archive(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_member_names(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def extract(self, member_name, dest_dir):
        raise NotImplementedError()

    def get_member(self, member_name):
        return ArchiveMember(self, member_name)

    def get_compressed_content_name(self, expected_extension):
        basename = os.path.basename(self.source.name)

        if not basename.endswith(expected_extension):
            return f"{basename}-content"

        return basename[: -len(expected_extension)]


class ArchiveMember(File):
    def __init__(self, container, member_name):
        super().__init__(container=container)
        self._name = member_name
        self._temp_dir = None
        self._path = None

    @property
    def path(self):
        if self._path is None:
            assert self._temp_dir is None
            self._temp_dir = get_temporary_directory(
                suffix=self.container.__class__.__name__
            )
            logger.debug(
                "Unpacking %s from %s to %s",
                self._name,
                self.container.source.name,
                self._temp_dir.name,
            )
            with profile("container_extract", self.container):
                self._path = self.container.extract(
                    self._name, self._temp_dir.name
                )
        return self._path

    def cleanup(self):
        if self._path is not None:
            self._path = None
        if self._temp_dir is not None:
            try:
                self._temp_dir.cleanup()
            except FileNotFoundError:
                pass
            self._temp_dir = None
        super().cleanup()

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def is_socket_or_fifo(self):
        return False


class MissingArchiveLikeObject(AbstractMissingType):
    def getnames(self):
        return []

    def list(self, *args, **kwargs):
        return ""

    def close(self):
        pass


class MissingArchive(Archive, AbstractMissingType):
    @property
    def source(self):
        return None

    def open_archive(self):
        return MissingArchiveLikeObject()

    def close_archive(self):
        pass

    def get_member_names(self):
        return []

    def extract(self, member_name, dest_dir):
        # should never be called
        raise NotImplementedError()

    def get_member(self, member_name):
        return MissingFile("/dev/null")

    # Be nice to gzip and the likes
    @property
    def path(self):
        return "/dev/null"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2016, 2018-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import re

from diffoscope.tools import python_module_missing, tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command

logger = logging.getLogger(__name__)

try:
    import PyPDF2

    try:
        # PyPDF 2.x
        from PyPDF2.errors import PdfReadError
    except ImportError:
        # PyPDF 1.x
        from PyPDF2.utils import PdfReadError

except ImportError:  # noqa
    python_module_missing("PyPDF2")
    PyPDF2 = None


class Pdftotext(Command):
    @tool_required("pdftotext")
    def cmdline(self):
        return ["pdftotext", self.path, "-"]


class Dumppdf(Command):
    @tool_required("dumppdf")
    def cmdline(self):
        return ["dumppdf", "-adt", self.path]


class PdfFile(File):
    DESCRIPTION = "PDF documents"
    FILE_TYPE_RE = re.compile(r"^PDF document\b")

    def compare_details(self, other, source=None):
        xs = []

        if PyPDF2 is None:
            self.add_comment(
                "Installing the 'PyPDF2' package may produce a better output."
            )
        else:
            difference = Difference.from_text(
                self.dump_pypdf2_metadata(self),
                self.dump_pypdf2_metadata(other),
                self.name,
                other.name,
            )
            if difference:
                difference.add_comment("Document info")
            xs.append(difference)

            difference = Difference.from_text(
                self.dump_pypdf2_annotations(self),
                self.dump_pypdf2_annotations(other),
                self.name,
                other.name,
            )
            if difference:
                difference.add_comment("Annotations")
            xs.append(difference)

        xs.append(Difference.from_operation(Pdftotext, self.path, other.path))

        # Don't include verbose dumppdf output unless we won't see any any
        # differences without it.
        if not any(xs):
            xs.append(
                Difference.from_operation(Dumppdf, self.path, other.path)
            )

        return xs

    def dump_pypdf2_metadata(self, file):
        try:
            pdf = PyPDF2.PdfFileReader(file.path)
            document_info = pdf.getDocumentInfo()

            if document_info is None:
                return ""

            xs = []
            for k, v in sorted(document_info.items()):
                xs.append("{}: {!r}".format(k.lstrip("/"), v))

            return "\n".join(xs)
        except PdfReadError as e:
            msg = f"Could not extract PyPDF2 metadata from {os.path.basename(file.name)}: {e}"
            self.add_comment(msg)
            logger.error(msg)
            return ""

    def dump_pypdf2_annotations(self, file):
        try:
            pdf = PyPDF2.PdfFileReader(file.path)

            xs = []
            for x in range(pdf.getNumPages()):
                page = pdf.getPage(x)

                try:
                    for annot in page["/Annots"]:
                        subtype = annot.getObject()["/Subtype"]
                        if subtype == "/Text":
                            xs.append(annot.getObject()["/Contents"])
                except:
                    pass

            return "\n".join(xs)
        except PdfReadError as e:
            msg = f"Could not extract PyPDF2 annotations from {os.path.basename(file.name)}: {e}"
            file.add_comment(msg)
            logger.error(msg)
            return ""



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging

from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference

from .utils.file import File

logger = logging.getLogger(__name__)


class Symlink(File):
    DESCRIPTION = "symlinks"

    @classmethod
    def recognizes(cls, file):
        return file.is_symlink()

    @property
    def symlink_destination(self):
        return os.readlink(self.name)

    def create_placeholder(self):
        with get_named_temporary_file("w+", delete=False) as f:
            f.write(f"destination: {self.symlink_destination}\n")
            f.flush()
            return f.name

    @property
    def path(self):
        if not hasattr(self, "_placeholder"):
            self._placeholder = self.create_placeholder()
        return self._placeholder

    def cleanup(self):
        if hasattr(self, "_placeholder"):
            try:
                os.remove(self._placeholder)
            except FileNotFoundError:
                pass
            del self._placeholder
        super().cleanup()

    def compare(self, other, source=None):
        with open(self.path) as my_content, open(other.path) as other_content:
            return Difference.from_text_readers(
                my_content,
                other_content,
                self.name,
                other.name,
                source=source,
                comment="symlink",
            )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import heapq
import io
import logging
import subprocess

from . import feeders
from .exc import RequiredToolNotFound
from .diff import diff, reverse_unified_diff, diff_split_lines
from .excludes import operation_excluded

logger = logging.getLogger(__name__)


class Difference:
    def __init__(
        self,
        path1,
        path2,
        source=None,
        comment=None,
        has_internal_linenos=False,
        details=None,
        visuals=None,
        unified_diff=None,
    ):
        self._unified_diff = unified_diff

        self._comments = []
        if comment:
            if type(comment) is list:
                self._comments.extend(comment)
            else:
                self._comments.append(comment)

        # Allow to override declared file paths, useful when comparing
        # tempfiles
        if source:
            if type(source) is list:
                self._source1, self._source2 = source
            else:
                self._source1 = source
                self._source2 = source
        else:
            self._source1 = path1
            self._source2 = path2

        # Ensure renderable types
        if not isinstance(self._source1, str):
            raise TypeError("path1/source[0] is not a string")
        if not isinstance(self._source2, str):
            raise TypeError("path2/source[1] is not a string")

        # Whether the unified_diff already contains line numbers inside itself
        self._has_internal_linenos = has_internal_linenos
        self._details = details or []
        self._visuals = visuals or []
        self._size_cache = None

    def __repr__(self):
        return (
            f"<Difference {self._source1} -- {self._source2} {self._details}>"
        )

    def map_lines(self, f_diff, f_comment):
        unified_diff = self.unified_diff
        if unified_diff is not None:
            unified_diff = "".join(map(f_diff, diff_split_lines(unified_diff)))
        return self.__class__(
            self.source1,
            self.source2,
            comment=[
                "".join(map(f_comment, diff_split_lines(comment)))
                for comment in self._comments
            ],
            has_internal_linenos=self.has_internal_linenos,
            details=self._details[:],
            visuals=self._visuals[:],
            unified_diff=unified_diff,
        )

    def fmap(self, f):
        return f(
            self.__class__(
                self.source1,
                self.source2,
                comment=self._comments[:],
                has_internal_linenos=self.has_internal_linenos,
                details=[d.fmap(f) for d in self._details],
                visuals=self._visuals[:],
                unified_diff=self.unified_diff,
            )
        )

    def _reverse_self(self):
        # assumes we're being called from get_reverse()
        if self._visuals:
            raise NotImplementedError(
                "_reverse_self on VisualDifference is not yet implemented"
            )

        unified_diff = None
        if self.unified_diff is not None:
            unified_diff = reverse_unified_diff(self.unified_diff)

        return self.__class__(
            self.source2,
            self.source1,
            comment=self._comments,  # already copied by fmap in get_reverse
            has_internal_linenos=self.has_internal_linenos,
            details=self._details,  # already reversed by fmap in get_reverse, no need to copy
            unified_diff=unified_diff,
        )

    def get_reverse(self):
        logger.debug("Reverse orig %s %s", self.source1, self.source2)
        return self.fmap(Difference._reverse_self)

    def equals(self, other):
        return self == other or (
            self.unified_diff == other.unified_diff
            and self.source1 == other.source1
            and self.source2 == other.source2
            and self._comments == other._comments
            and self.has_internal_linenos == other.has_internal_linenos
            and all(x.equals(y) for x, y in zip(self._details, other._details))
            and all(x.equals(y) for x, y in zip(self._visuals, other._visuals))
        )

    def size(self):
        if self._size_cache is None:
            self._size_cache = sum(
                d.size_self() for d in self.traverse_depth()
            )
        return self._size_cache

    def size_self(self):
        """Size, excluding children."""
        return (
            (len(self.unified_diff) if self.unified_diff else 0)
            + (len(self.source1) if self.source1 else 0)
            + (len(self.source2) if self.source2 else 0)
            + sum(map(len, self.comments))
            + sum(v.size() for v in self._visuals)
        )

    def has_visible_children(self):
        """
        Whether there are visible children.

        Useful for e.g. choosing whether to display [+]/[-] controls.
        """
        return (
            self._unified_diff is not None
            or self._comments
            or self._details
            or self._visuals
        )

    def traverse_depth(self, depth=-1):
        yield self
        if depth != 0:
            for d in self._details:
                yield from d.traverse_depth(depth - 1)

    def traverse_breadth(self, queue=None):
        queue = queue if queue is not None else [self]
        if queue:
            top = queue.pop(0)
            yield top
            queue.extend(top._details)
            yield from self.traverse_breadth(queue)

    def traverse_heapq(self, scorer, yield_score=False, queue=None):
        """Traverse the difference tree using a priority queue, where each node
        is scored according to a user-supplied function, and nodes with smaller
        scores are traversed first (after they have been added to the queue).

        The function `scorer` takes two arguments, a node to score and the
        score of its parent node (or None if there is no parent). It should
        return the score of the input node.
        """
        queue = queue if queue is not None else [(scorer(self, None), self)]
        while queue:
            val, top = heapq.heappop(queue)
            prune_descendants = yield ((top, val) if yield_score else top)
            if not prune_descendants:
                for d in top._details:
                    heapq.heappush(queue, (scorer(d, val), d))

    @staticmethod
    def from_feeder(
        feeder1, feeder2, path1, path2, source=None, comment=None, **kwargs
    ):
        try:
            unified_diff = diff(feeder1, feeder2)
            if not unified_diff:
                return None
            return Difference(
                path1,
                path2,
                source,
                comment,
                unified_diff=unified_diff,
                **kwargs,
            )
        except RequiredToolNotFound:
            difference = Difference(path1, path2, source)
            difference.add_comment("diff is not available")
            if comment:
                difference.add_comment(comment)
            return difference

    @staticmethod
    def from_text(content1, content2, *args, **kwargs):
        """
        Works for both bytes and str objects.
        """
        # Avoid spawning diff if buffers have same contents
        if content1 == content2:
            return None
        return Difference.from_feeder(
            feeders.from_text(content1),
            feeders.from_text(content2),
            *args,
            **kwargs,
        )

    @staticmethod
    def from_raw_readers(file1, file2, *args, **kwargs):
        return Difference.from_feeder(
            feeders.from_raw_reader(file1),
            feeders.from_raw_reader(file2),
            *args,
            **kwargs,
        )

    @staticmethod
    def from_text_readers(file1, file2, *args, **kwargs):
        return Difference.from_feeder(
            feeders.from_text_reader(file1),
            feeders.from_text_reader(file2),
            *args,
            **kwargs,
        )

    @staticmethod
    def from_operation(klass, path1, path2, *args, **kwargs):
        return Difference.from_operation_exc(
            klass, path1, path2, *args, **kwargs
        )[0]

    @staticmethod
    def from_operation_exc(klass, path1, path2, *args, **kwargs):
        operation_args = kwargs.pop("operation_args", [])
        ignore_returncodes = kwargs.pop("ignore_returncodes", ())

        def operation_and_feeder(path):
            operation = None
            if path == "/dev/null":
                feeder = feeders.empty()
            else:
                operation = klass(path, *operation_args)
                feeder = feeders.from_operation(operation)
                if operation_excluded(operation.full_name()):
                    return None, None, True
                operation.start()
            return feeder, operation, False

        feeder1, operation1, excluded1 = operation_and_feeder(path1)
        feeder2, operation2, excluded2 = operation_and_feeder(path2)
        if not feeder1 or not feeder2:
            assert excluded1 or excluded2
            return None, True

        if "source" not in kwargs:
            source_op = operation1 or operation2
            kwargs["source"] = source_op.full_name(truncate=120)

        try:
            short = kwargs.pop("short", False)
            # If the outputs are expected to be short, store them in memory
            # and do a direct comparison, and only spawn diff if needed.
            if short:
                memfile1 = io.BytesIO()
                feeder1(memfile1)
                memfile2 = io.BytesIO()
                feeder2(memfile2)
                bytes1 = memfile1.getbuffer().tobytes()
                bytes2 = memfile2.getbuffer().tobytes()
                # Check if the buffers are the same before invoking diff
                if bytes1 == bytes2:
                    return None, True
                difference = Difference.from_text(
                    bytes1, bytes2, path1, path2, *args, **kwargs
                )
            else:
                difference = Difference.from_feeder(
                    feeder1, feeder2, path1, path2, *args, **kwargs
                )
        except subprocess.CalledProcessError as exc:
            if exc.returncode in ignore_returncodes:
                return None, False
            raise

        if not difference:
            return None, False

        if (
            operation1
            and operation1.error_string
            and operation2
            and operation2.error_string
            and operation1.full_name() == operation2.full_name()
        ):
            # Output is the same, so don't repeat the output
            difference.add_comment(
                "error from `{}`:".format(operation1.full_name())
            )
            difference.add_comment(operation1.error_string)
        else:
            if operation1 and operation1.error_string:
                difference.add_comment(
                    "error from `{}` (a):".format(operation1.full_name())
                )
                difference.add_comment(operation1.error_string)
            if operation2 and operation2.error_string:
                difference.add_comment(
                    "error from `{}` (b):".format(operation2.full_name())
                )
                difference.add_comment(operation2.error_string)

        return difference, False

    @property
    def comment(self):
        return "\n".join(self._comments)

    @property
    def comments(self):
        return self._comments

    def add_comment(self, comment):
        for line in comment.splitlines():
            self._comments.append(line)
        self._size_cache = None

    @property
    def source1(self):
        return self._source1

    @property
    def source2(self):
        return self._source2

    @property
    def unified_diff(self):
        return self._unified_diff

    @unified_diff.setter
    def unified_diff(self, value):
        self._unified_diff = value

    @property
    def has_internal_linenos(self):
        return self._has_internal_linenos

    @property
    def details(self):
        return self._details

    @property
    def visuals(self):
        return self._visuals

    def add_details(self, differences):
        if len([d for d in differences if type(d) is not Difference]) > 0:
            raise TypeError("'differences' must contains Difference objects'")
        self._details.extend(differences)
        self._size_cache = None

    def add_visuals(self, visuals):
        if any([type(v) is not VisualDifference for v in visuals]):
            raise TypeError("'visuals' must contain VisualDifference objects'")
        self._visuals.extend(visuals)
        self._size_cache = None

    def has_ordering_differences_only(self):
        """
        Check if difference is only in line ordering.
        """

        if not self.unified_diff:
            return False

        diff_lines = self.unified_diff.splitlines()

        added_lines = [line[1:] for line in diff_lines if line.startswith("+")]
        removed_lines = [
            line[1:] for line in diff_lines if line.startswith("-")
        ]

        # Faster check: does number of lines match?
        if len(added_lines) != len(removed_lines):
            return False

        if added_lines == removed_lines:
            return False

        return sorted(added_lines) == sorted(removed_lines)

    def check_for_ordering_differences(self):
        if self.has_ordering_differences_only():
            self.add_comment("Ordering differences only")


class VisualDifference:
    def __init__(self, data_type, content, source):
        self._data_type = data_type
        self._content = content
        self._source = source

    @property
    def data_type(self):
        return self._data_type

    @property
    def content(self):
        return self._content

    @property
    def source(self):
        return self._source

    def size(self):
        return len(self.data_type) + len(self.content) + len(self.source)

    def equals(self, other):
        return self == other or (
            self._data_type == other._data_type
            and self._content == other._content
            and self._source == other._source
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging

from diffoscope.config import Config
from diffoscope.difference import Difference
from diffoscope.tools import python_module_missing

from .tar import TarContainer
from .utils.compare import compare_files
from .utils.file import File
from .utils.archive import ArchiveMember
from .utils.libarchive import LibarchiveContainer, list_libarchive
from .utils.specialize import specialize

try:
    from debian import deb822
except ImportError:
    python_module_missing("debian")
    deb822 = None

logger = logging.getLogger(__name__)


# Return a dict with build ids as keys and file as values for all deb in the
# given container
def get_build_id_map(container):
    d = {}
    for member_name, member in container.get_adjusted_members():
        # Let's assume the name will end with .deb to avoid looking at
        # too many irrelevant files
        if not member_name.endswith(".deb"):
            continue

        specialize(member)

        if isinstance(member, DebFile) and member.control:
            build_ids = member.control.get("Build-Ids", None)
            if build_ids:
                d.update({build_id: member for build_id in build_ids.split()})

    return d


class DebContainer(LibarchiveContainer):
    RE_DATA_TAR = re.compile(r"^data\.tar(\.gz|\.xz|\.zst|\.bz2|\.lzma)?$")
    RE_CONTROL_TAR = re.compile(r"^control\.tar(\.gz|\.xz|\.zst)?$")

    @property
    def data_tar(self):
        for name, member in self.get_adjusted_members():
            if not DebContainer.RE_DATA_TAR.match(name):
                continue

            specialize(member)

            if name.endswith(".tar"):
                return member

            return specialize(member.as_container.get_member("content"))

    @property
    def control_tar(self):
        for name, member in self.get_adjusted_members():
            if not DebContainer.RE_CONTROL_TAR.match(name):
                continue

            specialize(member)

            if name.endswith(".tar"):
                return member

            return specialize(member.as_container.get_member("content"))

    def perform_fuzzy_matching(self, my_members, other_members):
        matched = set()

        # Create local copies because they will be modified by consumer
        my_members = dict(my_members)
        other_members = dict(other_members)

        for name1 in my_members.keys():
            main, _ = os.path.splitext(name1)
            candidates = [
                name2
                for name2 in other_members.keys() - matched
                if os.path.splitext(name2)[0] == main
            ]
            if len(candidates) == 1:
                yield name1, candidates[0], 0
                matched.add(candidates[0])


class DebFile(File):
    CONTAINER_CLASSES = [DebContainer]
    FILE_TYPE_RE = re.compile(r"^Debian binary package")

    @property
    def md5sums(self):
        if not hasattr(self, "_md5sums"):
            control_tar = self.as_container.control_tar
            md5sums_file = (
                control_tar.as_container.lookup_file("./md5sums")
                if control_tar
                else None
            )
            if isinstance(md5sums_file, Md5sumsFile):
                self._md5sums = md5sums_file.parse()
            else:
                logger.debug("Unable to find a md5sums file")
                self._md5sums = {}
        return self._md5sums

    @property
    def control(self):
        if not deb822:
            return None

        if not hasattr(self, "_control"):
            control_file = (
                self.as_container.control_tar.as_container.lookup_file(
                    "./control"
                )
            )
            if control_file:
                with open(control_file.path, "rb") as f:
                    self._control = deb822.Deb822(f)

        return self._control

    def compare_details(self, other, source=None):
        return [
            Difference.from_text_readers(
                list_libarchive(self.path),
                list_libarchive(other.path),
                self.path,
                other.path,
                source="file list",
            )
        ]


class Md5sumsFile(File):
    @classmethod
    def recognizes(cls, file):
        return (
            isinstance(file, ArchiveMember)
            and file.name == "./md5sums"
            and isinstance(file.container.source, ArchiveMember)
            and isinstance(
                file.container.source.container.source, ArchiveMember
            )
            and DebContainer.RE_CONTROL_TAR.match(
                file.container.source.container.source.name
            )
            and isinstance(
                file.container.source.container.source.container.source,
                DebFile,
            )
        )

    def parse(self):
        try:
            md5sums = {}
            with open(self.path, "r", encoding="utf-8") as f:
                for line in f:
                    md5sum, path = re.split(r"\s+", line.strip(), maxsplit=1)
                    md5sums[f"./{path}"] = md5sum
            return md5sums
        except (UnicodeDecodeError, ValueError):
            logger.debug("Malformed md5sums, ignoring.")
            return {}

    def strip_checksum(self, path):
        with open(path, "r", encoding="utf-8") as f:
            for line in f:
                yield " ".join(line.split(" ")[2:])

    def compare_details(self, other, source=None):
        return [
            compare_files(
                self, other, source="md5sums", diff_content_only=True
            ),
            Difference.from_text_readers(
                self.strip_checksum(self.path),
                self.strip_checksum(other.path),
                self.path,
                other.path,
                source="line order",
            ),
        ]


class DebTarContainer(TarContainer):
    def comparisons(self, other):
        my_md5sums = {}
        other_md5sums = {}

        if self.source:
            my_md5sums = self.source.container.source.container.source.md5sums
        if other.source:
            other_md5sums = (
                other.source.container.source.container.source.md5sums
            )

        for my_member, other_member, comment in super().comparisons(other):
            if (
                not Config().force_details
                and my_member.name == other_member.name
                and my_md5sums.get(my_member.name, "my")
                == other_md5sums.get(other_member.name, "other")
            ):
                logger.debug("Skipping %s: identical md5sum", my_member.name)
                continue
            yield my_member, other_member, comment


class DebDataTarFile(File):
    CONTAINER_CLASSES = [DebTarContainer]

    @classmethod
    def recognizes(cls, file):
        return (
            isinstance(file, ArchiveMember)
            and isinstance(file.container.source, ArchiveMember)
            and DebContainer.RE_DATA_TAR.match(file.container.source.name)
            and isinstance(file.container.source.container.source, DebFile)
        )

    def compare_details(self, other, source=None):
        return [
            Difference.from_text_readers(
                list_libarchive(self.path, ignore_errors=True),
                list_libarchive(other.path, ignore_errors=True),
                self.path,
                other.path,
                source="file list",
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import stat
import logging

from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference

from .binary import FilesystemFile
from .utils.file import File

logger = logging.getLogger(__name__)


class Device(File):
    DESCRIPTION = "character/block devices"

    @classmethod
    def recognizes(cls, file):
        return file.is_device()

    def get_device(self):
        assert isinstance(self, FilesystemFile)
        st = os.lstat(self.name)
        return st.st_mode, os.major(st.st_rdev), os.minor(st.st_rdev)

    def has_same_content_as(self, other):
        try:
            return self.get_device() == other.get_device()
        except (AttributeError, OSError):
            # 'other' is not a device, or something.
            logger.debug("has_same_content: Not a device: %s", other)
            return False

    def create_placeholder(self):
        with get_named_temporary_file(mode="w+", delete=False) as f:
            f.write(format_device(*self.get_device()))
            f.flush()
            return f.name

    @property
    def path(self):
        if not hasattr(self, "_placeholder"):
            self._placeholder = self.create_placeholder()
        return self._placeholder

    def cleanup(self):
        if hasattr(self, "_placeholder"):
            os.remove(self._placeholder)
            del self._placeholder
        super().cleanup()

    def compare(self, other, source=None):
        with open(self.path) as my_content, open(other.path) as other_content:
            return Difference.from_text_readers(
                my_content,
                other_content,
                self.name,
                other.name,
                source=source,
                comment="device",
            )


def format_device(mode, major, minor):
    if stat.S_ISCHR(mode):
        kind = "character"
    elif stat.S_ISBLK(mode):
        kind = "block"
    else:
        kind = "weird"
    return f"device:{kind}\nmajor: {major}\nminor: {minor}\n"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016 Ximin Luo <infinity0@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import zlib
import os.path
import logging

from diffoscope.difference import Difference

from .utils.archive import Archive
from .utils.file import File

RLIB_BYTECODE_OBJECT_V1_DATASIZE_OFFSET = 15
RLIB_BYTECODE_OBJECT_V1_DATA_OFFSET = 23
ZLIB_DEFAULT_COMPRESSION = b"\x78\x9C"

logger = logging.getLogger(__name__)


class RustObjectContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".deflate")]

    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, member_name)
        logger.debug("rust-object extracting to %s", dest_path)
        # See librustc_trans/back/link.rs for details of this format
        with open(dest_path, "wb") as fpw, open(self.source.path, "rb") as fpr:
            raw_deflate = fpr.read()[RLIB_BYTECODE_OBJECT_V1_DATA_OFFSET:]
            # decompressobj() ignores the (non-existent) checksum; a zlib.decompress() would error
            raw_inflate = zlib.decompressobj().decompress(
                ZLIB_DEFAULT_COMPRESSION + raw_deflate
            )
            fpw.write(raw_inflate)
        return dest_path


class RustObjectFile(File):
    DESCRIPTION = "Rust object files (.deflate)"
    CONTAINER_CLASSES = [RustObjectContainer]
    FILE_TYPE_HEADER_PREFIX = b"RUST_OBJECT\x01\x00\x00\x00"
    FILE_EXTENSION_SUFFIX = {".deflate"}

    def compare_details(self, other, source=None):
        return [
            Difference.from_text(
                self.magic_file_type,
                other.magic_file_type,
                self,
                other,
                source="metadata",
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright ©      2015  Helmut Grohne <helmut@subdivi.de>
# Copyright © 2015-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import sys
import logging
import importlib
import traceback

from ..tools import python_module_missing
from ..logging import line_eraser


logger = logging.getLogger(__name__)


class ComparatorManager:
    COMPARATORS = (
        ("directory.Directory",),
        ("missing_file.MissingFile",),
        ("symlink.Symlink",),
        ("device.Device",),
        ("socket_or_fifo.SocketOrFIFO",),
        ("debian.DotChangesFile", "debian_fallback.DotChangesFile"),
        ("debian.DotDscFile", "debian_fallback.DotDscFile"),
        ("debian.DotBuildinfoFile", "debian_fallback.DotBuildinfoFile"),
        ("deb.Md5sumsFile",),
        ("deb.DebDataTarFile",),
        ("decompile.AsmFunction",),
        ("elf.ElfSection",),
        ("macho.MachoSection",),
        ("macho.MachoArchitecture",),
        ("binwalk.BinwalkFile",),
        ("ps.PsFile",),
        ("javascript.JavaScriptFile",),
        ("json.JSONFile",),
        ("xml.XMLFile",),
        ("xmlb.XMLBFile",),
        ("openssl.Pkcs7File",),
        ("openssl.MobileProvisionFile",),
        ("text.TextFile",),
        ("bzip2.Bzip2File",),
        ("cpio.CpioFile",),
        ("deb.DebFile",),
        ("hdf.Hdf5File",),
        ("dex.DexFile",),
        ("elf.ElfFile",),
        ("macho.MachoFile",),
        ("fsimage.FsImageFile",),
        ("llvm.LlvmBitCodeFile",),
        ("sqlite.Sqlite3Database",),
        ("wasm.WasmFile",),
        ("fonts.TtfFile",),
        ("fontconfig.FontconfigCacheFile",),
        ("gettext.MoFile",),
        ("ipk.IpkFile",),
        ("rust.RustObjectFile",),
        ("ffprobe.FfprobeFile",),
        ("gnumeric.GnumericFile",),
        ("gzip.GzipFile",),
        ("haskell.HiFile",),
        ("icc.IccFile",),
        ("iso9660.Iso9660File",),
        ("java.ClassFile",),
        ("lz4.Lz4File",),
        ("mono.MonoExeFile",),
        ("pdf.PdfFile",),
        ("png.PngFile",),
        ("ppu.PpuFile",),
        ("rdata.RdbFile",),
        ("rdata.RdsFile",),
        ("rpm.RpmFile", "rpm_fallback.RpmFile"),
        ("squashfs.SquashfsFile",),
        ("ar.ArFile",),
        ("tar.TarFile",),
        ("xz.XzFile",),
        ("apk.ApkFile",),
        ("odt.OdtFile",),
        ("ocaml.OcamlInterfaceFile",),
        ("docx.DocxFile",),
        ("zip.MozillaZipFile",),
        ("zip.JmodJavaModule",),
        ("zip.ZipFile",),
        ("image.JPEGImageFile",),
        ("image.ICOImageFile",),
        ("cbfs.CbfsFile",),
        ("git.GitIndexFile",),
        ("android.AndroidBootImgFile",),
        ("openssh.PublicKeyFile",),
        ("gif.GifFile",),
        ("pcap.PcapFile",),
        ("pe32.Pe32PlusFile",),
        ("pgp.PgpFile",),
        ("pgp.PgpSignature",),
        ("python.PycFile",),
        ("sphinx.SphinxInventoryFile",),
        ("kbx.KbxFile",),
        ("fit.FlattenedImageTreeFile",),
        ("dtb.DeviceTreeFile",),
        ("ogg.OggFile",),
        ("uimage.UimageFile",),
        ("xsb.XsbFile",),
        ("berkeley_db.BerkeleyDBFile",),
        ("zst.ZstFile",),
        ("vmlinuz.VmlinuzFile",),
    )

    _singleton = {}

    def __init__(self):
        self.__dict__ = self._singleton

        if not self._singleton:
            self.reload()

    def reload(self):
        self.classes = []

        for xs in self.COMPARATORS:
            errors = []
            for x in xs:
                package, klass_name = x.rsplit(".", 1)

                try:
                    mod = importlib.import_module(
                        f"diffoscope.comparators.{package}",
                    )
                except ModuleNotFoundError as e:
                    python_module_missing(e.name)
                    errors.append((x, e))
                    continue
                except ImportError as e:
                    errors.append((x, e))
                    continue

                self.classes.append(getattr(mod, klass_name))
                break
            else:  # noqa
                logger.error(
                    "Could not import {}{}".format(
                        "any of " if len(xs) > 1 else "", ", ".join(xs)
                    )
                )
                for x in errors:
                    logger.error("Original error for %s:", x[0])
                    sys.stderr.buffer.write(line_eraser())
                    traceback.print_exception(None, x[1], x[1].__traceback__)
                sys.exit(2)

        logger.debug("Loaded %d comparator classes", len(self.classes))

    def format_descriptions(self):
        def gen_descriptions():
            for x in self.classes:
                try:
                    yield x.DESCRIPTION
                except AttributeError:
                    pass

        xs = list(sorted(set(gen_descriptions()), key=str.upper))

        return "{} and {}.\n".format(", ".join(xs[:-1]), xs[-1])



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016, 2018-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from .utils.file import File


class AbstractRpmFile(File):
    FILE_TYPE_RE = re.compile(r"^RPM\s")


class RpmFile(AbstractRpmFile):
    DESCRIPTION = "RPM archives"

    def compare(self, other, source=None):
        difference = self.compare_bytes(other)
        if not difference:
            return None
        difference.add_comment(
            "Unable to import Python rpm module. Falling back to binary comparison."
        )
        return difference
